/* telnet.c
 * Dissector of telnet protocol
 *
 * $Id: $
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2009 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 * based on: ettercap -- dissector telnet --
 *   Copyright ALoR & NaGA. Web http://ettercap.sourceforge.net/
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <pcap.h>
#include <linux/tcp.h>

#include "proto.h"
#include "dmemory.h"
#include "etypes.h"
#include "flow.h"
#include "log.h"
#include "dnsdb.h"
#include "telnet.h"
#include "pei.h"

#define TELNET_TMP_DIR    "telnet"

/* info id */
static int ip_id;
static int ipv6_id;
static int ip_src_id;
static int ip_dst_id;
static int tcp_id;
static int port_src_id;
static int port_dst_id;
static int lost_id;
static int telnet_id;

/* pei id */
static int pei_host_id;
static int pei_user_id;
static int pei_password_id;
static int pei_cmd_id;

static volatile unsigned int incr;


static int TelnetSkipCommand(unsigned char *ptr, unsigned char *end)
{
    int len;
    
    len = 0;
    while(*ptr == 0xff && ptr < end) {
        /* sub option 0xff 0xfa ... ... 0xff 0xf0 */
        if (*(ptr + 1) == 0xfa) {
            ptr += 1;
            len++;
            /* search the sub-option end (0xff 0xf0) */
            do {
                ptr += 1;
                len++;
            } while(*ptr != 0xff && ptr < end);
            /* skip the sub-option end */
            ptr += 2;
            len += 2;
        }
        else {
            /* normal option 0xff 0xXX 0xXX */
            ptr += 3;
            len += 3;
        }
    }

    return len;
}


/* 
 * convert 0x00 char into spaces (0x20) so we can
 * use str*() functions on the buffer...
 */
static void TelnetConvertZeros(unsigned char *ptr, unsigned char *end)
{
    bool i;

    /* 
     * walk the entire buffer, but skip the last
     * char, if it is 0x00 it is actually the string
     * terminator
     */
    i = FALSE;
    while (ptr < end) {
        if (*ptr == 0x00) {
            /* convert the char to a space */
            if (i && *(ptr - 1) == '\r') {
                *ptr = '\n';
            }
            else {
                *ptr = ' ';
            }
        }
        ptr++;
        i = TRUE;
    }
}


static packet *TelnetDissector(int flow_id)
{
    packet *pkt;
    ftval lost;
    long offset;
    char cmd_file[TELNET_FILENAME_PATH_SIZE];
    FILE *fp;

    LogPrintf(LV_DEBUG, "Telnet id: %d", flow_id);

    /* init (this for each telnet stream) */
    sprintf(cmd_file, "%s/%s/telnet_%lu_%p_%i.txt", ProtTmpDir(), TELNET_TMP_DIR, time(NULL), cmd_file, incr);
    incr++;
    fp = fopen(cmd_file, "w");

    /* first packet */
    pkt = FlowGetPkt(flow_id);
    while (pkt != NULL) {
        offset = 0;
        /* check if there are packet lost */
        ProtGetAttr(pkt->stk, lost_id, &lost);
        //ProtStackFrmDisp(pkt->stk, TRUE); /* this function display the structure of packet stack of this packet */
        if (lost.uint8 == FALSE && pkt->len != 0) {
            /* no packet lost and packet with data */
            /* skip the telnet commands, we are interested only in readable data */
#warning "to do: reassemble telnet message from many tcp packets"
            offset = TelnetSkipCommand((unsigned char *)(pkt->data), (unsigned char *)(pkt->data + pkt->len));
            if (offset < pkt->len) {
                TelnetConvertZeros((unsigned char *)(pkt->data + offset), (unsigned char *)(pkt->data + pkt->len));
                fwrite(pkt->data + offset, 1, pkt->len - offset, fp);
            }
        }
        else if (lost.uint8) {
            fprintf(fp, "----->xplico: packets lost (size: %lub) <-----", pkt->len);
        }

        /* new packet */
        PktFree(pkt);
        pkt = FlowGetPkt(flow_id);
    }
    if (fp != NULL) {
        fclose(fp);
    }
    LogPrintf(LV_DEBUG, "Telnet... bye bye id: %d", flow_id);

    return NULL;
}


static bool TelnetVerifyCheck(int flow_id, bool check)
{
    packet *pkt;
    bool udp;
    bool ret;
    short cnt, cnt_lim;

    cnt = 0;
    udp = FALSE;
    ret = FALSE;
    pkt = FlowGetPktCp(flow_id);

    /* numer of packet to verify */
    if (FlowIsClose(flow_id) == TRUE)
        cnt_lim = FlowPktNum(flow_id);
    else
        cnt_lim = TELNET_PKT_LIMIT - 1;
    if (!check) {
        cnt_lim = 1;
    }
    
    if (pkt != NULL)
        PktFree(pkt);

    return !check;
}


static bool TelnetVerify(int flow_id)
{
    return TelnetVerifyCheck(flow_id, FALSE);
}


static bool TelnetCheck(int flow_id)
{
    return TelnetVerifyCheck(flow_id, TRUE);
}


int DissecRegist(void)
{
    proto_heury_dep hdep;
    proto_dep dep;
    pei_cmpt peic;

    memset(&hdep, 0, sizeof(proto_heury_dep));
    memset(&dep, 0, sizeof(proto_dep));
    memset(&peic, 0, sizeof(pei_cmpt));

    /* protocol name */
    ProtName("Telnet", "telnet");

    /* hdep: tcp */
    hdep.name = "tcp";
    hdep.ProtCheck = TelnetCheck;
    hdep.pktlim = TELNET_PKT_LIMIT;
    //ProtHeuDep(&hdep);

     /* dep: tcp */
    dep.name = "tcp";
    dep.attr = "tcp.dstport";
    dep.type = FT_UINT16;
    dep.val.uint16 = TCP_PORT_TELNET;
    dep.ProtCheck = TelnetVerify;
    dep.pktlim = TELNET_PKT_LIMIT;
    ProtDep(&dep);

    /* PEI components */
    peic.abbrev = "host";
    peic.desc = "Host name or IP";
    ProtPeiComponent(&peic);

    peic.abbrev = "user";
    peic.desc = "User name";
    ProtPeiComponent(&peic);
    
    peic.abbrev = "password";
    peic.desc = "Password";
    ProtPeiComponent(&peic);

    peic.abbrev = "cmd";
    peic.desc = "Commands";
    ProtPeiComponent(&peic);

    /* dissectors subdissectors registration */
    ProtDissectors(NULL, TelnetDissector, NULL, NULL);

    return 0;
}


int DissectInit(void)
{
    char telnet_dir[256];

    /* part of file name */
    incr = 0;

    /* info id */
    ip_id = ProtId("ip");
    ipv6_id = ProtId("ipv6");
    tcp_id = ProtId("tcp");
    ip_dst_id = ProtAttrId(ip_id, "ip.dst");
    ip_src_id = ProtAttrId(ip_id, "ip.src");
    port_dst_id = ProtAttrId(tcp_id, "tcp.dstport");
    port_src_id = ProtAttrId(tcp_id, "tcp.srcport");
    lost_id = ProtAttrId(tcp_id, "tcp.lost");
    telnet_id = ProtId("telnet");

    /* pei id */
    pei_host_id = ProtPeiComptId(telnet_id, "host");
    pei_user_id = ProtPeiComptId(telnet_id, "user");
    pei_password_id= ProtPeiComptId(telnet_id, "password");
    pei_cmd_id = ProtPeiComptId(telnet_id, "cmd");

    /* telnet tmp directory */
    sprintf(telnet_dir, "%s/%s", ProtTmpDir(), TELNET_TMP_DIR);
    mkdir(telnet_dir, 0x01FF);

    return 0;
}
